#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <math.h>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>

#include "multivariate_gaussian.h"
#include "common/dynamic.h"

unsigned int EF(arguments_len) = 1;
struct dynamic_type EF(arguments_descr)[1] = {
  {"dim", offsetof(struct family, dim), POSINT},
};


unsigned int EF(source_len) = 2;
struct dynamic_type EF(source_descr)[2] = {
  {"mu",    offsetof(struct source, mu)   , VECTOR},
  {"Sigma", offsetof(struct source, Sigma), MATRIX},
};

unsigned int EF(natural_len) = 2;
struct dynamic_type EF(natural_descr)[2] = {
  {"theta1", offsetof(struct natural, theta1), VECTOR},
  {"theta2", offsetof(struct natural, theta2), MATRIX},
};

unsigned int EF(expectation_len) = 2;
struct dynamic_type EF(expectation_descr)[2] = {
  {"eta1", offsetof(struct expectation, eta1), VECTOR},
  {"eta2", offsetof(struct expectation, eta2), MATRIX},
};

source* EF(create_source)(family* ef, gsl_vector* mu, gsl_matrix* Sigma) {
  source* param = EF(new_param)(ef, SOURCE);
  param->mu    = mu;
  param->Sigma = Sigma;

  return param;
}

natural* EF(create_natural)(family* ef, gsl_vector* theta1, gsl_matrix* theta2) {
  natural* param = EF(new_param)(ef, NATURAL);
  param->theta1 = theta1;
  param->theta2 = theta2;

  return param;
}

expectation* EF(create_expectation)(family* ef, gsl_vector* eta1, gsl_matrix* eta2) {
  expectation* param = EF(new_param)(ef, EXPECTATION);
  param->eta1 = eta1;
  param->eta2 = eta2;

  return param;
}

double EF(pdf)(family* ef, double* point, param* param0) {
  CHECK_EF(ef, param0);

  int status;

  gsl_vector_view input = gsl_vector_view_array(point, ef->dim);
  gsl_vector* x = &input.vector;
  
  source* lambda = (source*)EF(as_source)(ef, param0);

  gsl_vector* mu    = lambda->mu;
  gsl_matrix* Sigma = lambda->Sigma;

  /* Inspired by Scikit Multivariate Normal pdf for GMM (scikit-learn/sklearn/mixture/gmm.py) */

  /* (x-mu) */
  gsl_vector* x_minus_mu = gsl_vector_alloc(ef->dim);
  gsl_vector_memcpy(x_minus_mu, x);
  gsl_vector_sub(x_minus_mu, mu);

  /* Sigma = L L^T */
  gsl_matrix* L = gsl_matrix_alloc(ef->dim, ef->dim);
  gsl_matrix_memcpy(L, Sigma);
  status = gsl_linalg_cholesky_decomp(L);
  if (status) {
    GSL_ERROR_VAL("Non positive definite covariance matrix",
                  GSL_ERANGE, GSL_NAN);
  }

  /* sqrt(det(Sigma)) */
  /* gsl_linalg_LU_det computes the product of diagonal elements */
  double sqrt_det_Sigma = gsl_linalg_LU_det(L, 1);

  gsl_vector* tmp = gsl_vector_alloc(ef->dim);
  gsl_vector_memcpy(tmp, x_minus_mu);
  status = gsl_blas_dtrsv(CblasLower, CblasNoTrans, CblasNonUnit, L, tmp);
  if (status) {
    GSL_ERROR_VAL("Triangular solve failure",
                  GSL_ERANGE, GSL_NAN);
  }

  double x_minus_mu_inv_Sigma_x_minus_mu;
  status = gsl_blas_ddot(tmp, tmp, &x_minus_mu_inv_Sigma_x_minus_mu);
  if (status) {
    GSL_ERROR_VAL("Scalar product failure",
                  GSL_ERANGE, GSL_NAN);
  }

  double tmp1 = exp(-0.5 * x_minus_mu_inv_Sigma_x_minus_mu);
  double tmp2 = pow(2.0 * M_PI, ef->dim / 2.0) * sqrt_det_Sigma;
 
  /* free memory */
  gsl_vector_free(x_minus_mu);
  gsl_vector_free(tmp);
  gsl_matrix_free(L);
  ef->delete_param(ef, (param*)lambda);

  return tmp1 / tmp2;

  /* Scipy multivariate_normal code (scipy/stats/_multivariate.py) may be interesting */
}

void EF(lambda2theta)(family* ef, param* lambda0, param* theta0) {
  debug("lambda2theta\n");
  CHECK_EF(ef, lambda0);
  CHECK_EF(ef, theta0);
  CHECK_TYPE(lambda0, SOURCE);
  CHECK_TYPE(theta0, NATURAL);
  /* source lambda = (source)lambda0; */
  /* natural theta = (natural)theta0; */
}

void EF(theta2lambda)(family* ef, param* theta0, param* lambda0) {
  debug("theta2lambda\n");
  CHECK_EF(ef, lambda0);
  CHECK_EF(ef, theta0);
  CHECK_TYPE(lambda0, SOURCE);
  CHECK_TYPE(theta0, NATURAL);
  /* source lambda = (source)lambda0; */
  /* natural theta = (natural)theta0; */
}

void EF(lambda2eta)(family* ef, param* lambda0, param* eta0) {
  debug("lambda2eta\n");
  CHECK_EF(ef, lambda0);
  CHECK_EF(ef, eta0);
  CHECK_TYPE(lambda0, SOURCE);
  CHECK_TYPE(eta0, EXPECTATION);
  /* source lambda   = (source)lambda0; */
  /* expectation eta = (expectation)eta0; */
}

void EF(eta2lambda)(family* ef, param* eta0, param* lambda0) {
  debug("eta2lambda\n");
  CHECK_EF(ef, lambda0);
  CHECK_EF(ef, eta0);
  CHECK_TYPE(lambda0, SOURCE);
  CHECK_TYPE(eta0, EXPECTATION);
  source* lambda   = (source*)lambda0;
  expectation* eta = (expectation*)eta0;

  gsl_vector_memcpy(lambda->mu, eta->eta1);

  /* Sigma <- -1 * eta2 */
  gsl_matrix_memcpy(lambda->Sigma, eta->eta2);
  gsl_matrix_scale(lambda->Sigma, -1);

  /* Sigma <- - eta1 * eta1^T + Sigma */
  gsl_blas_dger(-1, eta->eta1, eta->eta1, lambda->Sigma);
}

double EF(F)(family* ef, param* theta0) {
  CHECK_EF(ef, theta0);
  CHECK_TYPE(theta0, NATURAL);
  /* natural theta = (natural)theta0; */
  /* double theta1 = theta->theta1; */
  /* double theta2 = theta->theta2; */

  return 0;
}

void EF(gradF)(family* ef, param* theta0, param* eta0) {
  CHECK_EF(ef, theta0);
  CHECK_EF(ef, eta0);
  CHECK_TYPE(theta0, NATURAL);
  CHECK_TYPE(eta0,   EXPECTATION);
  /* natural     theta = (natural)theta0; */
  /* expectation eta   = (expectation)eta0; */
}

double EF(G)(family* ef, param* eta0) {
  CHECK_EF(ef, eta0);
  CHECK_TYPE(eta0, EXPECTATION);
  /* expectation eta = (expectation)eta0; */
  /* double eta1 = eta->eta1; */
  /* double eta2 = eta->eta2; */

  return 0;
}

void EF(gradG)(family* ef, param* eta0, param* theta0) {
  CHECK_EF(ef, eta0);
  CHECK_EF(ef, theta0);
  CHECK_TYPE(eta0,   EXPECTATION);
  CHECK_TYPE(theta0, NATURAL);
  /* expectation eta   = (expectation)eta0; */
  /* natural     theta = (natural)theta0; */
}

void EF(t)(family* ef, double* x, param* eta0) {
  CHECK_EF(ef, eta0);
  CHECK_TYPE(eta0, EXPECTATION);
  expectation* eta = (expectation*)eta0;

  /* TODO use a vector view here !!! */
  /* eta1 <- x */
  for(unsigned int i=0; i<ef->dim; i++) {
    gsl_vector_set(eta->eta1, i, x[i]);
  }

  /* eta2 <- 0 */
  gsl_matrix_set_zero(eta->eta2);
  
  /* eta2 <- - x * x^T = - eta1 * eta1^T + eta2 */
  gsl_blas_dger(-1, eta->eta1, eta->eta1, eta->eta2);
}

double EF(k)(family* ef, double x) {
  return 0;
}

double* EF(rand)(family* ef, param* param0, gsl_rng* rng) {
  CHECK_EF(ef, param0);

  source* lambda = (source*)EF(as_source)(ef, param0);

  /* 1. cholskey(sigma) = A */
  /*    A = cholesky(Sigma) */
  gsl_matrix* A = gsl_matrix_alloc(ef->dim, ef->dim);
  gsl_matrix_memcpy(A, lambda->Sigma);
  gsl_linalg_cholesky_decomp(A);

  /* 2. generate uniform gaussian vector r */
  gsl_vector* r = gsl_vector_alloc(ef->dim);
  for(unsigned int i=0; i<ef->dim; i++) {
    gsl_vector_set(r, i, gsl_ran_gaussian(rng, 1.0));
  }
  
  /* 3. matrix vector triangular product with gsl_blas_dtrmv A * r */
  /*    r = A*r */
  gsl_blas_dtrmv(CblasLower, CblasNoTrans, CblasNonUnit, A, r);
    
  /* 4. add mu to Ar and that will be a vector of multivariate normal random numbers */
  /*    r <- r + mu */
  gsl_vector_add(r, lambda->mu);

  /* free memory */
  gsl_vector_free(r);
  gsl_matrix_free(A);
  
  return r->data;
}

#define CONSTRUCTOR
#include "common/all.c"

family* NAME(unsigned int dim) {
  family* ef = talloc(NULL, struct family);

  ef->name         = __S(NAME);
  ef->dim          = dim;
  ef->pdf          = &EF(pdf         );
  ef->iter_pdf     = &EF(iter_pdf    );
  ef->lambda2theta = &EF(lambda2theta);
  ef->theta2lambda = &EF(theta2lambda);
  ef->lambda2eta   = &EF(lambda2eta  );
  ef->eta2lambda   = &EF(eta2lambda  );
  ef->theta2eta    = &EF(theta2eta   );
  ef->eta2theta    = &EF(eta2theta   );
  ef->fprint       = &EF(fprint      );
  ef->F            = &EF(F           );
  ef->gradF        = &EF(gradF       );
  ef->G            = &EF(G           );
  ef->gradG        = &EF(gradG       );
  ef->mle          = &EF(mle         );
  ef->new_param    = &EF(new_param   );
  ef->delete_param = &EF(delete_param);
  ef->copy_param   = &EF(copy_param  );
  ef->t            = &EF(t           );
  ef->add          = &EF(add         );
  ef->scale        = &EF(scale       );
  ef->scalar       = &EF(scalar      );
  ef->zero         = &EF(zero        );
  ef->rand         = &EF(rand        );
  ef->as_source      = &EF(as_source     );
  ef->as_natural     = &EF(as_natural    );
  ef->as_expectation = &EF(as_expectation);

  return ef;
}

